import numpy
from primordial.equations import Equations as _Equations

class Equations(_Equations):
    def __init__(self, K, V):
        self.K = K
        self.V = V

    def __call__(self, t, y):
        N, phi, dphi = y
        H2 = self.H2(y)
        H = numpy.sqrt(H2)
        ddphi = -3*H*dphi - self.V.d(phi)
        return [H, dphi, ddphi]

    def H2(self, y):
        N, phi, dphi = y
        H2 = (dphi**2/2 + self.V(phi))/3 - self.K*numpy.exp(-2*N)
        return H2

    def inflating(self, y):
        N, phi, dphi = y
        return self.V(phi) - dphi**2 

    def sol(self, sol):
        sol.N, sol.phi, sol.dphi = sol.y
        sol.H = numpy.sqrt(self.H2(sol.y))
        return sol


def KD_initial_conditions(t0, K, N_p, phi_p): 
    b = K * numpy.exp(-2*N_p)
    return [
            N_p + numpy.log(t0)/3 - 9/14 * b * t0**(4./3),
            phi_p - numpy.sqrt(2./3)*numpy.log(t0) - 27*numpy.sqrt(6)/56*b*t0**(4./3),
            -numpy.sqrt(2./3)/t0 - 9*numpy.sqrt(6)/14 * b * t0**(1./3),
            ]



import numpy
from numpy.testing import assert_allclose, assert_almost_equal
from primordial.solver import Solver
from primordial.events import Inflation, Stationary
from primordial.equations_t import Equations, KD_initial_conditions
from primordial.potentials import ChaoticPotential

def test_Solver():
    for K in [-1, 0, +1]:
        m = 1
        V = ChaoticPotential(m)
        equations = Equations(K, V)

        events= [Inflation(equations, +1), 
                 Inflation(equations, -1, True), 
                 Stationary(equations, terminal=True)]

        solver = Solver(equations, events)

        N_p = -1.5
        phi_p = 23
        t = numpy.logspace(-5,10,10**6)
        y0 = KD_initial_conditions(t[0], K, N_p, phi_p)
        sol = solver.solve(t, y0)

        t = 1e-5
        sol = solver.solve(t, y0)



class Equations(object):
    def sol(self, sol):
        return sol

class Event(object):
    def __init__(self, equations, direction=0, terminal=False):
        self.equations = equations
        self.direction = direction
        self.terminal = terminal



class Potential(object):
    pass

class ChaoticPotential(Potential):
    def __init__(self, m):
        self.m = m

    def __call__(self, phi):
        return self.m**2*phi**2/2

    def d(self, phi):
        return self.m**2*phi



from scipy import integrate

class Solver(object):
    method = 'RK45'

    def __init__(self, equations, events=None):
        self.equations = equations
        self.events = events

    def solve(self, t, y0):
        try:
            t0 = t[0]
            t1 = t[-1]
            t_eval = t
        except TypeError:
            t0 = t
            t1 = 1e100
            t_eval = None
        sol = integrate.solve_ivp(self.equations, (t0, t1), y0, 
                                  method=self.method, t_eval=t_eval, events=self.events)
        return self.equations.sol(sol)

    def events(self):
        return None



class Event(object):
    def __init__(self, equations, direction=0, terminal=False):
        self.equations = equations
        self.direction = direction
        self.terminal = terminal


class Inflation(Event):
    def __call__(self, t, y):
        return self.equations.inflating(y)


class Stationary(Event):
    def __call__(self, t, y):
        return self.equations.H2(y)

